"""Setup package."""

import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="OpenAlchemy",
    version="2.0.1",
    author="David Andersson",
    author_email="anderssonpublic@gmail.com",
    description="Maps an OpenAPI schema to SQLAlchemy models.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/jdkandersson/OpenAlchemy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Topic :: Internet",
        "Typing :: Typed",
    ],
    python_requires=">=3.7",
    install_requires=[
        "SQLAlchemy>=1.0",
        "jsonschema>=3",
        "Jinja2>=2",
        "sqlalchemy-stubs>=0.3",
    ],
    include_package_data=True,
    extras_require={
        "yaml": ["PyYAML"],
        "wheel": ["wheel"],
        "dev": [
            "pylint",
            "mypy",
            "pydocstyle",
            "black",
            "pre-commit",
            "isort<5.0.0",
            "Sphinx",
            "doc8",
            "connexion[swagger-ui]",
            "Flask-SQLAlchemy",
            "alembic",
            "rstcheck",
        ],
        "test": [
            "bandit",
            "pytest",
            "pytest-cov",
            "pytest-flake8",
            "pytest-flask",
            "pytest-flask-sqlalchemy",
            "pytest-randomly",
            "PyYAML",
            "connexion[swagger-ui]",
            "typeguard",
            "sqlalchemy_mixins",
            "wheel",
        ],
        ":python_version<'3.8'": ["typing_extensions>=3.7.4"],
    },
    entry_points={
        "console_scripts": ["openalchemy=open_alchemy.cli:main"],
    },
)



"""Pre-processor that defines any foreign keys."""

import typing

from .. import helpers as oa_helpers
from .. import types
from . import helpers


def _requires_foreign_key(schemas: types.Schemas, schema: types.Schema) -> bool:
    """
    Check whether the property requires a foreign key to be defined.

    Assume schema is valid.

    Foreign keys are required for many-to-one, one-to-one and one-to-many relationships.
    The following rules capture this:
    1. relationship property that is not
    2. a many-to-many relationship.

    Args:
        schemas: All the defined schemas used to resolve any $ref.
        schema: The schema of the property.

    Returns:
        Whether the property requires a foreign key.

    """
    # Filter for relationship properties
    property_type = oa_helpers.property_.calculate_type(schemas=schemas, schema=schema)
    if property_type != types.PropertyType.RELATIONSHIP:
        return False

    # Filter for not many-to-many relationship
    relationship_type = oa_helpers.relationship.calculate_type(
        schema=schema, schemas=schemas
    )
    if relationship_type == types.RelationshipType.MANY_TO_MANY:
        return False
    return True


def _foreign_key_property_not_defined(
    schemas: types.Schema,
    parent_schema: types.Schema,
    property_name: str,
    property_schema: types.Schema,
) -> bool:
    """
    Check whether the foreign key property is not already defined.

    Assume the full relationship schema is valid.

    Args:
        schemas: All defined schemas used to resolve any $ref.
        parent_schema: The schema that contains the relationship property.
        property_name: The name of the property.
        property_schema: The schema of the property.

    Returns:
        Whether the foreign key property is not already defined.

    """
    # Retrieve the property name
    type_ = oa_helpers.relationship.calculate_type(
        schema=property_schema, schemas=schemas
    )
    column_name = oa_helpers.foreign_key.calculate_column_name(
        type_=type_,
        property_schema=property_schema,
        schemas=schemas,
    )
    target_schema = oa_helpers.foreign_key.get_target_schema(
        type_=type_,
        parent_schema=parent_schema,
        property_schema=property_schema,
        schemas=schemas,
    )
    foreign_key_property_name = oa_helpers.foreign_key.calculate_prop_name(
        type_=type_,
        column_name=column_name,
        property_name=property_name,
        target_schema=target_schema,
        schemas=schemas,
    )

    # Look for the foreign key property name on the schema the foreign key needs to be
    # defined on
    modify_schema = oa_helpers.foreign_key.get_modify_schema(
        type_=type_,
        parent_schema=parent_schema,
        property_schema=property_schema,
        schemas=schemas,
    )
    properties = helpers.iterate.properties_items(schema=modify_schema, schemas=schemas)
    property_names = map(lambda arg: arg[0], properties)
    contains_foreign_key_property_name = any(
        filter(lambda name: name == foreign_key_property_name, property_names)
    )
    if contains_foreign_key_property_name:
        return False
    return True


class TArtifacts(helpers.process.TArtifacts):
    """The return value of _calculate_schema."""

    property_schema: types.ColumnSchema


def _calculate_foreign_key_property_artifacts(
    schemas: types.Schema,
    parent_name: str,
    parent_schema: types.Schema,
    property_name: str,
    property_schema: types.Schema,
) -> TArtifacts:
    """
    Calculate the artifacts for the schema for the foreign key property.

    Assume the full relationship schema is valid.
    Assume that the relationship is not many-to-many

    Args:
        schemas: All defined schemas used to resolve any $ref.
        parent_name: The name of the schema that contains the relationship property.
        parent_schema: The schema that contains the relationship property.
        property_name: The name of the property.
        property_schema: The schema of the property.

    Returns:
        The name of the schema to store the property onto and the name and schema of the
        foreign key property.

    """
    # Retrieve the schema of the property that is targeted by the foreign key
    relationship_type = oa_helpers.relationship.calculate_type(
        schema=property_schema, schemas=schemas
    )
    assert relationship_type != types.RelationshipType.MANY_TO_MANY

    column_name = oa_helpers.foreign_key.calculate_column_name(
        type_=relationship_type,
        property_schema=property_schema,
        schemas=schemas,
    )
    target_schema = oa_helpers.foreign_key.get_target_schema(
        type_=relationship_type,
        parent_schema=parent_schema,
        property_schema=property_schema,
        schemas=schemas,
    )
    target_schema_properties = helpers.iterate.properties_items(
        schema=target_schema, schemas=schemas
    )
    foreign_key_target = next(
        filter(lambda arg: arg[0] == column_name, target_schema_properties), None
    )
    assert foreign_key_target is not None
    _, foreign_key_target_schema = foreign_key_target

    # Check whether the property is required
    required: typing.Optional[bool] = None
    if relationship_type != types.RelationshipType.ONE_TO_MANY:
        required_items = helpers.iterate.required_items(
            schema=parent_schema, schemas=schemas
        )
        required = any(filter(lambda name: name == property_name, required_items))
    # Calculate nullable for all but one-to-many relationships based on property
    nullable: typing.Optional[bool] = None
    relationship_type = oa_helpers.relationship.calculate_type(
        schema=property_schema, schemas=schemas
    )
    if relationship_type != types.RelationshipType.ONE_TO_MANY:
        nullable = oa_helpers.peek.nullable(schema=property_schema, schemas=schemas)
    default = oa_helpers.peek.default(schema=foreign_key_target_schema, schemas=schemas)
    server_default = oa_helpers.peek.server_default(
        schema=foreign_key_target_schema, schemas=schemas
    )
    nullable = oa_helpers.calculate_nullable(
        nullable=nullable,
        generated=False,
        defaulted=default is not None or server_default is not None,
        required=required,
    )

    # Retrieve information about the foreign key schema
    foreign_key = oa_helpers.foreign_key.calculate_foreign_key(
        column_name=column_name,
        target_schema=target_schema,
        schemas=schemas,
    )
    property_type = oa_helpers.peek.type_(
        schema=foreign_key_target_schema, schemas=schemas
    )
    format_ = oa_helpers.peek.format_(schema=foreign_key_target_schema, schemas=schemas)
    max_length = oa_helpers.peek.max_length(
        schema=foreign_key_target_schema, schemas=schemas
    )

    # Calculate the schema
    foreign_key_property_schema: types.ColumnSchema = {
        "type": property_type,
        "x-dict-ignore": True,
        "nullable": nullable,
        "x-foreign-key": foreign_key,
    }
    if format_ is not None:
        foreign_key_property_schema["format"] = format_
    if max_length is not None:
        foreign_key_property_schema["maxLength"] = max_length
    if default is not None:
        foreign_key_property_schema["default"] = default
    if server_default is not None:
        foreign_key_property_schema["x-server-default"] = server_default

    # Calculate other artifacts
    modify_name = oa_helpers.foreign_key.get_modify_name(
        type_=relationship_type,
        parent_name=parent_name,
        property_schema=property_schema,
        schemas=schemas,
    )
    foreign_key_property_name = oa_helpers.foreign_key.calculate_prop_name(
        type_=relationship_type,
        column_name=column_name,
        property_name=property_name,
        target_schema=target_schema,
        schemas=schemas,
    )

    return TArtifacts(
        modify_name, foreign_key_property_name, foreign_key_property_schema
    )


def _get_schema_foreign_keys(
    schemas: types.Schemas,
    schema_name: str,
    schema: types.Schema,
) -> helpers.process.TArtifactsIter:
    """
    Retrieve the foreign keys for a schema.

    Assume schema is constructable.

    Algorithm:
    1. validate property schema and filter for whether a foreign key is required,
    2. validate the full relationship schema and filter for foreign keys not already
        defined and
    3. capture the artifacts for the foreign key.

    Args:
        schemas: All schemas.
        schema_name: The name of the schema.
        schema: A constructable schema.

    Returns:
        Iterable with all foreign keys of the constructable schema.

    """
    # Get all the properties of the schema
    names_properties = helpers.iterate.properties_items(
        schema=schema, schemas=schemas, stay_within_model=True
    )
    # Validate and remove properties that don't require foreign keys
    foreign_key_name_properties = filter(
        lambda args: _requires_foreign_key(schemas, args[1]), names_properties
    )
    # Validate relationship and remove properties that already have a defined foreign
    # key
    undefined_foreign_key_name_properties = filter(
        lambda args: _foreign_key_property_not_defined(
            schemas, schema, args[0], args[1]
        ),
        foreign_key_name_properties,
    )
    # Convert to artifacts
    return map(
        lambda args: _calculate_foreign_key_property_artifacts(
            schemas, schema_name, schema, args[0], args[1]
        ),
        undefined_foreign_key_name_properties,
    )


def _foreign_keys_to_schema(
    foreign_keys: helpers.process.TArtifactsIter,
) -> types.Schema:
    """
    Convert to the schema with the foreign keys from foreign key artifacts.

    Args:
        foreign_keys: The foreign keys to convert.

    Returns:
        The schema with the foreign keys.

    """
    return {
        "type": "object",
        "properties": {
            property_name: schema for _, property_name, schema in foreign_keys
        },
    }


def process(*, schemas: types.Schemas):
    """
    Pre-process the schemas to add foreign keys as required.

    Args:
        schemas: The schemas to process.

    """
    # Retrieve foreign keys
    foreign_keys = helpers.process.get_artifacts(
        schemas=schemas, get_schema_artifacts=_get_schema_foreign_keys
    )
    # Map to a schema for each grouped foreign keys
    foreign_key_schemas = helpers.process.calculate_outputs(
        artifacts=foreign_keys, calculate_output=_foreign_keys_to_schema
    )
    # Convert to list to resolve iterator
    foreign_key_schema_list = list(foreign_key_schemas)
    # Add foreign keys to schemas
    for name, foreign_key_schema in foreign_key_schema_list:
        schemas[name] = {"allOf": [schemas[name], foreign_key_schema]}



"""Types shared across modules."""

import dataclasses
import datetime
import enum
import typing

try:  # pragma: no cover
    from typing import Literal  # pylint: disable=unused-import
    from typing import Protocol
    from typing import TypedDict
except ImportError:  # pragma: no cover
    from typing_extensions import Literal  # type: ignore
    from typing_extensions import Protocol  # type: ignore
    from typing_extensions import TypedDict  # type: ignore

Schema = typing.Dict[str, typing.Any]
Schemas = typing.Dict[str, Schema]
AllOfSpec = typing.List[Schema]
TKwargs = typing.Dict[str, typing.Any]
TOptKwargs = typing.Optional[TKwargs]


class ModelFactory(Protocol):
    """Defines interface for model factory."""

    def __call__(self, *, name: str) -> typing.Type:
        """Call signature for ModelFactory."""
        ...


class GetBase(Protocol):
    """Defines interface for the get_base function."""

    def __call__(self, *, name: str, schemas: Schemas) -> typing.Type:
        """Call signature for get_base."""
        ...


# Unique consraint types
ColumnList = typing.List[str]
ColumnListList = typing.List[ColumnList]


class _UniqueBase(TypedDict, total=True):
    """Base class for unique schema."""

    columns: typing.List[str]


class Unique(_UniqueBase, total=False):
    """Unique schema."""

    name: typing.Optional[str]


UniqueList = typing.List[Unique]
AnyUnique = typing.Union[ColumnList, ColumnListList, Unique, UniqueList]
# Index types


class _IndexBase(TypedDict, total=True):
    """Base class for index schema."""

    expressions: typing.List[str]


class Index(_IndexBase, total=False):
    """Index schema."""

    name: typing.Optional[str]
    unique: bool


IndexList = typing.List[Index]
AnyIndex = typing.Union[ColumnList, ColumnListList, Index, IndexList]
# Type for the default value
TColumnDefault = typing.Optional[typing.Union[str, int, float, bool]]
# Type for the default value expressed in Python
TPyColumnDefault = typing.Optional[
    typing.Union[str, int, float, bool, bytes, datetime.date, datetime.datetime]
]


_ColumnSchemaBase = TypedDict(
    "_ColumnSchemaBase",
    {
        "x-dict-ignore": bool,
        "format": str,
        "x-primary-key": bool,
        "maxLength": int,
        "nullable": bool,
        "description": str,
        "x-json": bool,
        "default": TColumnDefault,
        "x-server-default": str,
        "x-generated": bool,
        "readOnly": bool,
        "writeOnly": bool,
        "x-foreign-key": str,
    },
    total=False,
)


class ColumnSchema(_ColumnSchemaBase, total=True):
    """Schema for column definitions."""

    type: str


_ObjectRefSchemaBase = TypedDict(
    "_ObjectRefSchemaBase", {"type": str, "x-de-$ref": str}, total=True
)


class ObjectRefSchema(_ObjectRefSchemaBase, total=False):
    """Schema for object reference definitions."""

    nullable: bool
    description: str
    readOnly: bool
    writeOnly: bool


_ArrayRefSchemaBase = TypedDict(
    "_ArrayRefSchemaBase",
    {"description": str, "readOnly": bool, "writeOnly": bool},
    total=False,
)


class ArrayRefSchema(_ArrayRefSchemaBase, total=True):
    """Schema for array reference definitions."""

    type: str
    items: _ObjectRefSchemaBase


class TNameSchema(typing.NamedTuple):
    """The name and schema of a schema."""

    name: str
    schema: Schema


@enum.unique
class PropertyType(str, enum.Enum):
    """The type of a property."""

    SIMPLE = "SIMPLE"
    JSON = "JSON"
    RELATIONSHIP = "RELATIONSHIP"
    BACKREF = "BACKREF"


@enum.unique
class RelationshipType(str, enum.Enum):
    """The relationship type."""

    MANY_TO_ONE = "MANY_TO_ONE"
    ONE_TO_ONE = "ONE_TO_ONE"
    ONE_TO_MANY = "ONE_TO_MANY"
    MANY_TO_MANY = "MANY_TO_MANY"


TMixins = typing.List[str]


@dataclasses.dataclass
class PropertyArtifacts:
    """Information about a property."""

    type: Literal[
        PropertyType.SIMPLE,
        PropertyType.JSON,
        PropertyType.RELATIONSHIP,
        PropertyType.BACKREF,
    ]
    schema: typing.Union[
        Schema,
        ColumnSchema,
        ObjectRefSchema,
        ArrayRefSchema,
    ]
    required: typing.Optional[bool]
    description: typing.Optional[str]


class _OpenApiSimplePropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the OpenAPI artifacts for a simple property."""

    format: str
    max_length: int
    nullable: bool

    default: typing.Union[int, float, str, bool]

    read_only: bool
    write_only: bool


class OpenApiSimplePropertyTypedDict(_OpenApiSimplePropertyTypedDictBase, total=True):
    """TypedDict representation of the OpenAPI artifacts for a simple property."""

    type: str


@dataclasses.dataclass
class OpenApiSimplePropertyArtifacts:
    """OpenAPI artifacts for the simple property."""

    type: str
    format: typing.Optional[str]
    max_length: typing.Optional[int]
    nullable: typing.Optional[bool]

    default: typing.Optional[typing.Union[int, float, str, bool]]

    read_only: typing.Optional[bool]
    write_only: typing.Optional[bool]

    def to_dict(self) -> OpenApiSimplePropertyTypedDict:
        """Convert to dictionary."""
        return_dict: OpenApiSimplePropertyTypedDict = {"type": self.type}

        opt_keys: typing.List[
            Literal[
                "format",
                "max_length",
                "nullable",
                "default",
                "read_only",
                "write_only",
            ]
        ] = [
            "format",
            "max_length",
            "nullable",
            "default",
            "read_only",
            "write_only",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


class _ExtensionSimplePropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the extension artifacts for a simple property."""

    autoincrement: bool
    index: bool
    unique: bool
    server_default: str

    foreign_key: str

    kwargs: TKwargs
    foreign_key_kwargs: TKwargs


class ExtensionSimplePropertyTypedDict(
    _ExtensionSimplePropertyTypedDictBase, total=True
):
    """TypedDict representation of the extension artifacts for a simple property."""

    primary_key: bool


@dataclasses.dataclass
class ExtensionSimplePropertyArtifacts:
    """OpenAPI artifacts for the simple property."""

    primary_key: bool
    autoincrement: typing.Optional[bool]
    index: typing.Optional[bool]
    unique: typing.Optional[bool]
    server_default: typing.Optional[str]

    foreign_key: typing.Optional[str]

    kwargs: typing.Optional[TKwargs]
    foreign_key_kwargs: typing.Optional[TKwargs]

    dict_ignore: typing.Optional[bool]

    def to_dict(self) -> ExtensionSimplePropertyTypedDict:
        """Convert to dictionary."""
        return_dict: ExtensionSimplePropertyTypedDict = {
            "primary_key": self.primary_key
        }

        opt_keys: typing.List[
            Literal[
                "autoincrement",
                "index",
                "unique",
                "server_default",
                "foreign_key",
                "kwargs",
                "foreign_key_kwargs",
            ]
        ] = [
            "autoincrement",
            "index",
            "unique",
            "server_default",
            "foreign_key",
            "kwargs",
            "foreign_key_kwargs",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


class _SimplePropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the simple property."""

    description: str


class SimplePropertyTypedDict(_SimplePropertyTypedDictBase, total=True):
    """TypedDict representation of the simple property."""

    type: str
    open_api: OpenApiSimplePropertyTypedDict
    extension: ExtensionSimplePropertyTypedDict
    schema: ColumnSchema
    required: bool


@dataclasses.dataclass
class SimplePropertyArtifacts(PropertyArtifacts):
    """Information about a simple property."""

    type: Literal[PropertyType.SIMPLE]
    open_api: OpenApiSimplePropertyArtifacts
    extension: ExtensionSimplePropertyArtifacts
    schema: ColumnSchema
    required: bool

    def to_dict(self) -> SimplePropertyTypedDict:
        """Convert to dictionary."""
        return_dict: SimplePropertyTypedDict = {
            "type": self.type,
            "open_api": self.open_api.to_dict(),
            "extension": self.extension.to_dict(),
            "schema": self.schema,
            "required": self.required,
        }

        if self.description is not None:
            return_dict["description"] = self.description

        return return_dict


class OpenApiJsonPropertyTypedDict(TypedDict, total=False):
    """TypedDict representation of the OpenAPI artifacts for a JSON property."""

    nullable: bool

    read_only: bool
    write_only: bool


@dataclasses.dataclass
class OpenApiJsonPropertyArtifacts:
    """OpenAPI artifacts for the JSON property."""

    nullable: typing.Optional[bool]

    read_only: typing.Optional[bool]
    write_only: typing.Optional[bool]

    def to_dict(self) -> OpenApiJsonPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: OpenApiJsonPropertyTypedDict = {}

        opt_keys: typing.List[Literal["nullable", "read_only", "write_only"]] = [
            "nullable",
            "read_only",
            "write_only",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


class _ExtensionJsonPropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the extension artifacts for a JSON property."""

    index: bool
    unique: bool

    foreign_key: str

    kwargs: TKwargs
    foreign_key_kwargs: TKwargs


class ExtensionJsonPropertyTypedDict(_ExtensionJsonPropertyTypedDictBase, total=True):
    """TypedDict representation of the extension artifacts for a JSON property."""

    primary_key: bool


@dataclasses.dataclass
class ExtensionJsonPropertyArtifacts:
    """OpenAPI artifacts for the JSON property."""

    primary_key: bool
    index: typing.Optional[bool]
    unique: typing.Optional[bool]

    foreign_key: typing.Optional[str]

    kwargs: typing.Optional[TKwargs]
    foreign_key_kwargs: typing.Optional[TKwargs]

    def to_dict(self) -> ExtensionJsonPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: ExtensionJsonPropertyTypedDict = {"primary_key": self.primary_key}

        opt_keys: typing.List[
            Literal[
                "index",
                "unique",
                "foreign_key",
                "kwargs",
                "foreign_key_kwargs",
            ]
        ] = [
            "index",
            "unique",
            "foreign_key",
            "kwargs",
            "foreign_key_kwargs",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


class _JsonPropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the JSON property."""

    description: str


class JsonPropertyTypedDict(_JsonPropertyTypedDictBase, total=True):
    """TypedDict representation of the JSON property."""

    type: str
    open_api: OpenApiJsonPropertyTypedDict
    extension: ExtensionJsonPropertyTypedDict
    schema: Schema
    required: bool


@dataclasses.dataclass
class JsonPropertyArtifacts(PropertyArtifacts):
    """Information about a JSON property."""

    type: Literal[PropertyType.JSON]
    open_api: OpenApiJsonPropertyArtifacts
    extension: ExtensionJsonPropertyArtifacts
    schema: Schema
    required: bool

    def to_dict(self) -> JsonPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: JsonPropertyTypedDict = {
            "type": self.type,
            "open_api": self.open_api.to_dict(),
            "extension": self.extension.to_dict(),
            "schema": self.schema,
            "required": self.required,
        }

        if self.description is not None:
            return_dict["description"] = self.description

        return return_dict


@dataclasses.dataclass
class RelationshipPropertyArtifacts(PropertyArtifacts):
    """Information about a relationship property."""

    type: Literal[PropertyType.RELATIONSHIP]
    sub_type: Literal[
        RelationshipType.MANY_TO_ONE,
        RelationshipType.ONE_TO_ONE,
        RelationshipType.ONE_TO_MANY,
        RelationshipType.MANY_TO_MANY,
    ]
    parent: str
    backref_property: typing.Optional[str]
    kwargs: typing.Optional[TKwargs]
    write_only: typing.Optional[bool]
    description: typing.Optional[str]
    required: bool


@dataclasses.dataclass
class NotManyToManyRelationshipPropertyArtifacts(RelationshipPropertyArtifacts):
    """Information about a relationship that is not many-to-many property."""

    sub_type: Literal[
        RelationshipType.MANY_TO_ONE,
        RelationshipType.ONE_TO_ONE,
        RelationshipType.ONE_TO_MANY,
    ]
    foreign_key: str
    foreign_key_property: str


class _OneToManyRelationshipPropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the one-to-many relationship property."""

    backref_property: str
    kwargs: TKwargs
    write_only: bool
    description: str


class OneToManyRelationshipPropertyTypedDict(
    _OneToManyRelationshipPropertyTypedDictBase, total=True
):
    """TypedDict representation of the one-to-many relationship property."""

    type: str
    sub_type: str
    parent: str
    required: bool
    schema: ArrayRefSchema
    foreign_key: str
    foreign_key_property: str


@dataclasses.dataclass
class OneToManyRelationshipPropertyArtifacts(
    NotManyToManyRelationshipPropertyArtifacts
):
    """Information about a one-to-many relationship property."""

    sub_type: Literal[
        RelationshipType.ONE_TO_MANY,
    ]
    schema: ArrayRefSchema

    def to_dict(self) -> OneToManyRelationshipPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: OneToManyRelationshipPropertyTypedDict = {
            "type": self.type,
            "sub_type": self.sub_type,
            "parent": self.parent,
            "required": self.required,
            "schema": self.schema,
            "foreign_key": self.foreign_key,
            "foreign_key_property": self.foreign_key_property,
        }

        opt_keys: typing.List[
            Literal["backref_property", "kwargs", "write_only", "description"]
        ] = [
            "backref_property",
            "kwargs",
            "write_only",
            "description",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


class _XToOneRelationshipPropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the x-to-one relationship property."""

    backref_property: str
    kwargs: TKwargs
    write_only: bool
    description: str
    nullable: bool


class XToOneRelationshipPropertyTypedDict(
    _XToOneRelationshipPropertyTypedDictBase, total=True
):
    """TypedDict representation of the x-to-one relationship property."""

    type: str
    sub_type: str
    parent: str
    required: bool
    schema: ObjectRefSchema
    foreign_key: str
    foreign_key_property: str


@dataclasses.dataclass
class XToOneRelationshipPropertyArtifacts(NotManyToManyRelationshipPropertyArtifacts):
    """Information about a x-to-one relationship property."""

    sub_type: Literal[
        RelationshipType.MANY_TO_ONE,
        RelationshipType.ONE_TO_ONE,
    ]
    schema: ObjectRefSchema
    nullable: typing.Optional[bool]

    def to_dict(self) -> XToOneRelationshipPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: XToOneRelationshipPropertyTypedDict = {
            "type": self.type,
            "sub_type": self.sub_type,
            "parent": self.parent,
            "required": self.required,
            "schema": self.schema,
            "foreign_key": self.foreign_key,
            "foreign_key_property": self.foreign_key_property,
        }

        opt_keys: typing.List[
            Literal[
                "backref_property",
                "kwargs",
                "write_only",
                "description",
                "nullable",
            ]
        ] = [
            "backref_property",
            "kwargs",
            "write_only",
            "description",
            "nullable",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


ManyToOneRelationshipPropertyTypedDict = XToOneRelationshipPropertyTypedDict


@dataclasses.dataclass
class ManyToOneRelationshipPropertyArtifacts(XToOneRelationshipPropertyArtifacts):
    """Information about a many-to-one relationship property."""

    sub_type: Literal[
        RelationshipType.MANY_TO_ONE,
    ]

    def to_dict(  # pylint: disable=useless-super-delegation
        self,
    ) -> ManyToOneRelationshipPropertyTypedDict:
        """Convert to dictionary."""
        return super().to_dict()


OneToOneRelationshipPropertyTypedDict = XToOneRelationshipPropertyTypedDict


@dataclasses.dataclass
class OneToOneRelationshipPropertyArtifacts(XToOneRelationshipPropertyArtifacts):
    """Information about a one-to-one relationship property."""

    sub_type: Literal[
        RelationshipType.ONE_TO_ONE,
    ]

    def to_dict(  # pylint: disable=useless-super-delegation
        self,
    ) -> OneToOneRelationshipPropertyTypedDict:
        """Convert to dictionary."""
        return super().to_dict()


class _ManyToManyRelationshipPropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the many-to-many relationship property."""

    backref_property: str
    kwargs: TKwargs
    write_only: bool
    description: str


class ManyToManyRelationshipPropertyTypedDict(
    _ManyToManyRelationshipPropertyTypedDictBase, total=True
):
    """TypedDict representation of the many-to-many relationship property."""

    type: str
    sub_type: str
    parent: str
    required: bool
    schema: ArrayRefSchema
    secondary: str


@dataclasses.dataclass
class ManyToManyRelationshipPropertyArtifacts(RelationshipPropertyArtifacts):
    """Information about a x-to-one relationship property."""

    sub_type: Literal[
        RelationshipType.MANY_TO_MANY,
    ]
    secondary: str
    schema: ArrayRefSchema

    def to_dict(self) -> ManyToManyRelationshipPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: ManyToManyRelationshipPropertyTypedDict = {
            "type": self.type,
            "sub_type": self.sub_type,
            "parent": self.parent,
            "required": self.required,
            "schema": self.schema,
            "secondary": self.secondary,
        }

        opt_keys: typing.List[
            Literal["backref_property", "kwargs", "write_only", "description"]
        ] = [
            "backref_property",
            "kwargs",
            "write_only",
            "description",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


TAnyRelationshipPropertyTypedDict = typing.Union[
    ManyToOneRelationshipPropertyTypedDict,
    OneToOneRelationshipPropertyTypedDict,
    OneToManyRelationshipPropertyTypedDict,
    ManyToManyRelationshipPropertyTypedDict,
]
TAnyRelationshipPropertyArtifacts = typing.Union[
    ManyToOneRelationshipPropertyArtifacts,
    OneToOneRelationshipPropertyArtifacts,
    OneToManyRelationshipPropertyArtifacts,
    ManyToManyRelationshipPropertyArtifacts,
]


class _BackrefPropertyTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the backref property."""

    description: str


class BackrefPropertyTypedDict(_BackrefPropertyTypedDictBase, total=True):
    """TypedDict representation of the backref property."""

    type: str
    sub_type: str
    properties: typing.List[str]
    schema: Schema


@enum.unique
class BackrefSubType(str, enum.Enum):
    """The possible types of backreferences."""

    OBJECT = "OBJECT"
    ARRAY = "ARRAY"


@dataclasses.dataclass
class BackrefPropertyArtifacts(PropertyArtifacts):
    """Information about a back reference property."""

    type: Literal[PropertyType.BACKREF]
    sub_type: Literal[BackrefSubType.OBJECT, BackrefSubType.ARRAY]
    properties: typing.List[str]
    schema: Schema
    required: None

    def to_dict(self) -> BackrefPropertyTypedDict:
        """Convert to dictionary."""
        return_dict: BackrefPropertyTypedDict = {
            "type": self.type,
            "sub_type": self.sub_type,
            "properties": self.properties,
            "schema": self.schema,
        }

        if self.description is not None:
            return_dict["description"] = self.description

        return return_dict


TAnyPropertyTypedDict = typing.Union[
    SimplePropertyTypedDict,
    JsonPropertyTypedDict,
    TAnyRelationshipPropertyTypedDict,
    BackrefPropertyTypedDict,
]
TAnyPropertyArtifacts = typing.Union[
    SimplePropertyArtifacts,
    JsonPropertyArtifacts,
    TAnyRelationshipPropertyArtifacts,
    BackrefPropertyArtifacts,
]


class PropertyValue(TypedDict, total=True):
    """Artifacts for a property."""

    artifacts: TAnyPropertyTypedDict


PropertiesValue = typing.Dict[str, PropertyValue]


class _ModelTypedDictBase(TypedDict, total=False):
    """TypedDict representation of the model artifacts."""

    inherits: bool
    parent: str

    description: str

    mixins: TMixins

    kwargs: TKwargs

    composite_index: IndexList
    composite_unique: UniqueList


class ModelTypedDict(_ModelTypedDictBase, total=True):
    """TypedDict representation of the model artifacts."""

    tablename: str


class _ModelValueBase(TypedDict, total=False):
    """Record artifacts of a model."""

    properties: PropertiesValue


class ModelValue(_ModelValueBase, total=True):
    """Record artifacts of a model."""

    artifacts: ModelTypedDict


class ModelBackrefArtifacts(typing.NamedTuple):
    """Artifacts for model back references."""

    type: BackrefSubType
    child: str


@dataclasses.dataclass
class ModelExPropertiesArtifacts:
    """Information about a model excluding its properties."""

    tablename: str
    inherits: typing.Optional[bool]
    parent: typing.Optional[str]

    description: typing.Optional[str]

    mixins: typing.Optional[TMixins]

    kwargs: typing.Optional[TKwargs]

    composite_index: typing.Optional[IndexList]
    composite_unique: typing.Optional[UniqueList]

    backrefs: typing.List[typing.Tuple[str, ModelBackrefArtifacts]]

    def to_dict(self) -> ModelTypedDict:
        """Convert to dictionary."""
        return_dict: ModelTypedDict = {"tablename": self.tablename}

        opt_keys: typing.List[
            Literal[
                "inherits",
                "parent",
                "description",
                "mixins",
                "kwargs",
                "composite_index",
                "composite_unique",
            ]
        ] = [
            "inherits",
            "parent",
            "description",
            "mixins",
            "kwargs",
            "composite_index",
            "composite_unique",
        ]
        for opt_key in opt_keys:
            value = getattr(self, opt_key)
            if value is None:
                continue

            return_dict[opt_key] = value

        return return_dict


@dataclasses.dataclass
class ModelArtifacts(ModelExPropertiesArtifacts):
    """Full information about a model."""

    properties: typing.List[typing.Tuple[str, TAnyPropertyArtifacts]]


ModelsValue = typing.Dict[str, ModelValue]


class SpecValue(TypedDict, total=False):
    """Record artifacts for a specification."""

    models: ModelsValue


ModelsModelArtifacts = typing.Dict[str, ModelArtifacts]



"""Helper functions for relationships."""

import typing

from .. import types
from . import peek
from . import property_
from . import ref


def calculate_type(
    *, schema: types.Schema, schemas: types.Schemas
) -> types.RelationshipType:
    """
    Calculate the type of the relationship the schema defines.

    Assume that the property is a relationship.
    Assume that the property schema has been verified.

    Args:
        schema: The schema of the property.
        schemas: All defined schemas used to resolve any $ref.

    Returns:
        The type of relationship that is defined.

    """
    # Calculate the type of the property
    type_ = peek.type_(schema=schema, schemas=schemas)

    # Handle object
    if type_ == "object":
        # Retrieve uselist
        uselist = peek.prefer_local(
            get_value=peek.uselist, schema=schema, schemas=schemas
        )
        if uselist is False:
            return types.RelationshipType.ONE_TO_ONE
        return types.RelationshipType.MANY_TO_ONE

    # Retrieve the items schema
    items_schema = peek.items(schema=schema, schemas=schemas)
    assert items_schema is not None
    secondary = peek.secondary(schema=items_schema, schemas=schemas)
    if secondary is not None:
        return types.RelationshipType.MANY_TO_MANY
    return types.RelationshipType.ONE_TO_MANY


def get_ref_schema_many_to_x(
    *, property_schema: types.Schema, schemas: types.Schemas
) -> typing.Tuple[str, types.Schema]:
    """
    Get the schema referenced by a many-to-x relationship property.

    Assume the schema and schemas are valid.
    Assume the property schema is a many-to-x relationship property.

    Args:
        property_schema: The schema of the many-to-x relationship property.
        schemas: All defined schemas.

    Returns:
        The schema referenced in the many-to-x relationship.

    """
    items_schema = peek.items(schema=property_schema, schemas=schemas)
    assert items_schema is not None
    ref_value = peek.ref(schema=items_schema, schemas=schemas)
    assert ref_value is not None
    return ref.get_ref(ref=ref_value, schemas=schemas)


def is_relationship_type(
    *, type_: types.RelationshipType, schema: types.Schema, schemas: types.Schemas
) -> bool:
    """
    Check whether a property is a relationship of a particular type.

    Args:
        type_: The type of relationship to look for.
        schema: The schema of the property to check.
        schemas: All defined schemas.

    Returns:
        Whether the property is a relationship of a particular type.

    """
    property_type = property_.calculate_type(schema=schema, schemas=schemas)
    if property_type != types.PropertyType.RELATIONSHIP:
        return False

    relationship_type = calculate_type(schema=schema, schemas=schemas)
    return relationship_type == type_



"""Map an OpenAPI schema to SQLAlchemy models."""

import functools
import sys
import types as py_types
import typing

from sqlalchemy.ext import declarative

from open_alchemy import types as oa_types

from . import build as _build_module
from . import exceptions
from . import helpers as _helpers
from . import model_factory as _model_factory
from . import models_file as _models_file
from . import schemas as _schemas_module
from .build import PackageFormat

models = py_types.ModuleType("models")  # pylint: disable=invalid-name
sys.modules["open_alchemy.models"] = models


def init_model_factory(
    *,
    base: typing.Type,
    spec: oa_types.Schema,
    models_filename: typing.Optional[str] = None,
    spec_path: typing.Optional[str] = None,
) -> oa_types.ModelFactory:
    """
    Create factory that generates SQLAlchemy models based on OpenAPI specification.

    Args:
        base: The declarative base for the models.
        spec: The OpenAPI specification in the form of a dictionary.
        models_filename: The name of the file to write the models typing information to.
        spec_path: The path the the OpenAPI specification. Mainly used to support remote
            references.

    Returns:
        A factory that returns SQLAlchemy models derived from the base based on the
        OpenAPI specification.

    """
    # Record the spec path
    if spec_path is not None:
        _helpers.ref.set_context(path=spec_path)

    # Retrieving the schema from the specification
    if "components" not in spec:
        raise exceptions.MalformedSpecificationError(
            '"components" is a required key in the specification.'
        )
    components = spec.get("components", {})
    if "schemas" not in components:
        raise exceptions.MalformedSpecificationError(
            '"schemas" is a required key in the components of the specification.'
        )
    schemas = components.get("schemas", {})

    # Pre-processing schemas
    _schemas_module.process(schemas=schemas)

    # Getting artifacts
    schemas_artifacts = _schemas_module.artifacts.get_from_schemas(
        schemas=schemas, stay_within_model=True
    )

    # Binding the base and schemas
    bound_model_factories = functools.partial(
        _model_factory.model_factory,
        schemas=schemas,
        artifacts=schemas_artifacts,
        get_base=_get_base,
    )
    # Caching calls
    cached_model_factories = functools.lru_cache(maxsize=None)(bound_model_factories)

    # Making Base importable
    setattr(models, "Base", base)

    # Intercept factory calls to make models available
    def _register_model(*, name: str) -> typing.Type:
        """Intercept calls to model factory and register model on models."""
        model = cached_model_factories(name=name)
        setattr(models, name, model)
        return model

    if models_filename is not None:
        schemas_artifacts = _schemas_module.artifacts.get_from_schemas(
            schemas=schemas, stay_within_model=False
        )
        models_file_contents = _models_file.generate(artifacts=schemas_artifacts)
        with open(models_filename, "w") as out_file:
            out_file.write(models_file_contents)

    _helpers.define_all(model_factory=_register_model, schemas=schemas)

    return _register_model


BaseAndModelFactory = typing.Tuple[typing.Type, oa_types.ModelFactory]


def _init_optional_base(
    *,
    base: typing.Optional[typing.Type],
    spec: oa_types.Schema,
    models_filename: typing.Optional[str] = None,
    spec_path: typing.Optional[str] = None,
) -> BaseAndModelFactory:
    """Wrap init_model_factory with optional base."""
    if base is None:
        base = declarative.declarative_base()
    return (
        base,
        init_model_factory(
            base=base,
            spec=spec,
            models_filename=models_filename,
            spec_path=spec_path,
        ),
    )


def init_json(
    spec_filename: str,
    *,
    base: typing.Optional[typing.Type] = None,
    models_filename: typing.Optional[str] = None,
) -> BaseAndModelFactory:
    """
    Create SQLAlchemy models factory based on an OpenAPI specification as a JSON file.

    Args:
        spec_filename: filename of an OpenAPI spec in JSON format
        base: (optional) The declarative base for the models.
              If base=None, construct a new SQLAlchemy declarative base.
        models_filename: (optional) The path to write the models file to. If it is not
            provided, the models file is not created.

    Returns:
        A tuple (Base, model_factory), where:

        Base: a SQLAlchemy declarative base class
        model_factory: A factory that returns SQLAlchemy models derived from the
            base based on the OpenAPI specification.

    """
    # Most OpenAPI specs are YAML, so, for efficiency, we only import json if we
    # need it:
    import json  # pylint: disable=import-outside-toplevel

    with open(spec_filename) as spec_file:
        spec = json.load(spec_file)

    return _init_optional_base(
        base=base,
        spec=spec,
        models_filename=models_filename,
        spec_path=spec_filename,
    )


def init_yaml(
    spec_filename: str,
    *,
    base: typing.Optional[typing.Type] = None,
    models_filename: typing.Optional[str] = None,
) -> BaseAndModelFactory:
    """
    Create SQLAlchemy models factory based on an OpenAPI specification as a YAML file.

    Raise ImportError if pyyaml has not been installed.

    Args:
        spec_filename: filename of an OpenAPI spec in YAML format
        base: (optional) The declarative base for the models.
              If base=None, construct a new SQLAlchemy declarative base.
        models_filename: (optional) The path to write the models file to. If it is not
            provided, the models file is not created.

    Returns:
        A tuple (Base, model_factory), where:

        Base: a SQLAlchemy declarative base class
        model_factory: A factory that returns SQLAlchemy models derived from the
            base based on the OpenAPI specification.

    """
    try:
        import yaml  # pylint: disable=import-outside-toplevel
    except ImportError as exc:
        raise ImportError(
            "Using init_yaml requires the pyyaml package. Try `pip install pyyaml`."
        ) from exc

    with open(spec_filename) as spec_file:
        spec = yaml.load(spec_file, Loader=yaml.SafeLoader)

    return _init_optional_base(
        base=base,
        spec=spec,
        models_filename=models_filename,
        spec_path=spec_filename,
    )


def _get_base(*, name: str, schemas: oa_types.Schemas) -> typing.Type:
    """
    Retrieve the base class of a schema considering inheritance.

    If x-inherits is True, retrieve the parent. If it is a string, verify that the
    parent is valid. In either case, the model for that schema is used as the base
    instead of the usual base.
    If x-inherits is not present or False, return the usual base.

    Raise InheritanceConstructionOrderError if the parent of the schema has not been
    constructed when attempting to construct the child.

    Args:
        name: The name of the schema to determine the base for.
        schemas: All the schemas.

    Returns:
        The base of the model. Either the usual base or the model parent in the case of
        inheritance.

    """
    schema = schemas.get(name)
    if schema is None:
        raise exceptions.SchemaNotFoundError(f"Could not fund schema {name}.")

    if _helpers.schema.inherits(schema=schema, schemas=schemas):
        parent = _helpers.inheritance.retrieve_parent(schema=schema, schemas=schemas)
        try:
            return getattr(models, parent)
        except AttributeError as exc:
            raise exceptions.InheritanceError(
                "Any parents of a schema must be constructed before the schema can be "
                "constructed."
            ) from exc
    return getattr(models, "Base")


def build_json(
    spec_filename: str,
    package_name: str,
    dist_path: str,
    format_: PackageFormat = PackageFormat.NONE,
) -> None:
    """
    Create an OpenAlchemy distribution package with the SQLAlchemy models.

    The package can be uploaded to, for example, PyPI or a private repository for
    distribution.

    The formats can be combined with the bitwise operator or (``|``), for
    instance, building both sdist and wheel packages can be specified like that:

    .. code-block: python

        format_ = PackageFormat.SDIST|PackageFormat.WHEEL

    Args:
        spec_filename: filename of an OpenAPI spec in JSON format
        package_name: The name of the package.
        dist_path: The directory to output the package to.
        format_: (optional) The format(s) of the archive(s) to build.

    """
    # Most OpenAPI specs are YAML, so, for efficiency, we only import json if we
    # need it:
    import json  # pylint: disable=import-outside-toplevel

    with open(spec_filename) as spec_file:
        spec = json.load(spec_file)

    return _build_module.execute(
        spec=spec, name=package_name, path=dist_path, format_=format_
    )


def build_yaml(
    spec_filename: str,
    package_name: str,
    dist_path: str,
    format_: PackageFormat = PackageFormat.NONE,
) -> None:
    """
    Create an OpenAlchemy distribution package with the SQLAlchemy models.

    The package can be uploaded to, for example, PyPI or a private repository for
    distribution.

    The formats can be combined with the bitwise operator "or" (``|``), for
    instance, building both sdist and wheel packages can be specified like that:

    .. code-block:: python

        format_=PackageFormat.SDIST|PackageFormat.WHEEL

    Args:
        spec_filename: filename of an OpenAPI spec in YAML format
        package_name: The name of the package.
        dist_path: The directory to output the package to.
        format_: (optional) The format(s) of the archive(s) to build.

    """
    try:
        import yaml  # pylint: disable=import-outside-toplevel
    except ImportError as exc:
        raise ImportError(
            "Using init_yaml requires the pyyaml package. Try `pip install pyyaml`."
        ) from exc

    with open(spec_filename) as spec_file:
        spec = yaml.load(spec_file, Loader=yaml.SafeLoader)

    return _build_module.execute(
        spec=spec, name=package_name, path=dist_path, format_=format_
    )


__all__ = [
    "init_model_factory",
    "init_json",
    "init_yaml",
    "build_json",
    "build_yaml",
    "PackageFormat",
]



"""Generate models files based on individual models."""

import os
import sys
import typing

import jinja2

_DIRECTORY = os.path.dirname(__file__)
_TEMPLATE_FILE = os.path.join(_DIRECTORY, "template.j2")
with open(_TEMPLATE_FILE) as in_file:
    _TEMPLATE = in_file.read()

_ALL_IMPORTS = {"datetime", "typing"}


def generate(*, models: typing.List[str]) -> str:
    """
    Generate the models file.

    Args:
        models: The models to add to the models file.

    Returns:
        The source for the models file.

    """
    imports: typing.Set[str] = {"typing"}
    for model in models:
        if imports == _ALL_IMPORTS:
            break
        if "datetime." in model:
            imports.add("datetime")

    template = jinja2.Template(_TEMPLATE, trim_blocks=True)
    return template.render(
        imports=sorted(list(imports)),
        models=models,
        python_minor_version=sys.version_info[1],
    )



"""Convert simple types (not object nor array)."""

import datetime

from ... import exceptions
from ... import helpers
from ... import types as oa_types
from .. import types


def convert(
    value: types.TOptSimpleCol, *, schema: oa_types.Schema
) -> types.TOptSimpleDict:
    """
    Convert values with basic types to dictionary values.

    Raises InvalidInstanceError if the value is not of the type implied by the schema.

    Args:
        value: The value to convert.
        schema: The schema for the value.

    Returns:
        The value converted to the expected dictionary value.

    """
    type_ = helpers.peek.type_(schema=schema, schemas={})
    if value is None:
        return None

    if type_ == "integer":
        if not isinstance(value, int):
            raise exceptions.InvalidInstanceError(
                "Integer type columns must have int values."
            )
        return value
    if type_ == "number":
        if not isinstance(value, float):
            raise exceptions.InvalidInstanceError(
                "Number type columns must have float values."
            )
        return value
    if type_ == "string":
        return _handle_string(value, schema=schema)
    if type_ == "boolean":
        if not isinstance(value, bool):
            raise exceptions.InvalidInstanceError(
                "Boolean type columns must have bool values."
            )
        return value

    raise exceptions.FeatureNotImplementedError(f"Type {type_} is not supported.")


def _handle_string(value: types.TSimpleCol, *, schema: oa_types.Schema) -> str:
    """
    Convert string type column to str.

    Raises InvalidInstanceError if the value is not of the type implied by the schema.

    Args:
        value: The value to convert.

    Returns:
        The converted value.

    """
    format_ = helpers.peek.format_(schema=schema, schemas={})
    if format_ == "date":
        if not isinstance(value, datetime.date):
            raise exceptions.InvalidInstanceError(
                "String type columns with date format must have date values."
            )
        return value.isoformat()
    if format_ == "date-time":
        if not isinstance(value, datetime.datetime):
            raise exceptions.InvalidInstanceError(
                "String type columns with date-time format must have datetime "
                "values."
            )
        return value.isoformat()
    if format_ == "binary":
        if not isinstance(value, bytes):
            raise exceptions.InvalidInstanceError(
                "String type columns with binary format must have bytes values."
            )
        return value.decode()
    if not isinstance(value, str):
        raise exceptions.InvalidInstanceError(
            "String type columns must have str values."
        )
    return value



"""Helpers that clean up."""

from ... import types


def extension(*, schema: types.Schema) -> None:
    """
    Remove extension properties from a schema.

    Args:
        schema: The schema to clean.

    """
    extension_keys = list(filter(lambda key: key.startswith("x-"), schema.keys()))
    for key in extension_keys:
        del schema[key]



"""Read the value of an extension property, validate the schema and return it."""

import json
import os
import typing

from open_alchemy import exceptions
from open_alchemy import facades
from open_alchemy import types

_DIRECTORY = os.path.dirname(__file__)
_SCHEMAS_FILE = os.path.join(_DIRECTORY, "extension-schemas.json")
_COMMON_SCHEMAS_FILE = os.path.join(_DIRECTORY, "common-schemas.json")
_resolver, (_SCHEMAS, _) = facades.jsonschema.resolver(  # pylint: disable=invalid-name
    _SCHEMAS_FILE, _COMMON_SCHEMAS_FILE
)


def get(
    *,
    source: typing.Union[
        types.Schema,
        types.ColumnSchema,
        types.ObjectRefSchema,
        types.ArrayRefSchema,
    ],
    name: str,
    default: typing.Optional[typing.Any] = None,
    pop: bool = False,
) -> typing.Optional[typing.Any]:
    """
    Read the value of an extension property, validate the schema and return it.

    Raise MalformedExtensionPropertyError when the schema of the extension property is
    malformed.

    Args:
        source: The object to get the extension property from.
        name: The name of the property.
        default: The default value.
        pop: Whether to remove the value from the dictionary.

    Returns:
        The value of the property or the default value if it does not exist.

    """
    # Check for presence of name
    if name not in source:
        return default

    # Retrieve value
    value = source.get(name)
    if value is None:
        raise exceptions.MalformedExtensionPropertyError(
            f"The value of the {name} extension property cannot be null."
        )

    schema = _SCHEMAS.get(name)
    try:
        facades.jsonschema.validate(instance=value, schema=schema, resolver=_resolver)
    except facades.jsonschema.ValidationError as exc:
        raise exceptions.MalformedExtensionPropertyError(
            f"The value of the {json.dumps(name)} extension property is not "
            "valid. "
            f"The expected schema is {json.dumps(schema)}. "
            f"The given value is {json.dumps(value)}."
        ) from exc
    if pop:
        del source[name]  # type: ignore
    return value


def get_kwargs(
    *,
    source: typing.Union[
        types.Schema,
        types.ColumnSchema,
        types.ObjectRefSchema,
        types.ArrayRefSchema,
    ],
    reserved: typing.Optional[typing.Set[str]] = None,
    default: typing.Optional[typing.Any] = None,
    pop: bool = False,
    name: str = "x-kwargs",
) -> types.TOptKwargs:
    """
    Read the value of x-kwargs, validate the schema and return it.

    Raise MalformedExtensionPropertyError when the schema of the extension property is
        malformed.
    Raise MalformedExtensionPropertyError when the keys of the kwargs are not a string.
    Raise MalformedExtensionPropertyError if any keys of x-kwargs are in the reserved
        keys.

    Args:
        source: The object to get the extension property from.
        reserved: The reserved keys that should not be in the kwargs.
        default: The default value.
        pop: Whether to remove the value from the dictionary.
        name: The name of the extension property with the kwargs.

    Returns:
        The value of the property or the default value if it does not exist.

    """
    value = get(source=source, name=name, default=default, pop=pop)
    if value is None:
        return None

    # Check for dictionary to make mypy happy, in reality always passes
    if not isinstance(value, dict):  # pragma: no cover
        raise exceptions.MalformedExtensionPropertyError(
            "The value of x-kwargs must be an object."
        )

    # Check keys are strings
    if any(not isinstance(key, str) for key in value.keys()):
        raise exceptions.MalformedExtensionPropertyError(
            "The keys of x-kwargs must be strings."
        )

    # Check whether any reserved keys are in use
    if reserved is not None:
        if not reserved.isdisjoint(value.keys()):
            raise exceptions.MalformedExtensionPropertyError(
                "Some of the keys in x-kwargs refer to arguments that OpenAlchemy "
                "handles. To make use of the arguments, please use the relevant "
                "extension property. The following keys need to be removed: "
                f"{reserved.intersection(value.keys())}."
            )

    return value



"""Build a package with the OpenAlchemy models."""

import dataclasses
import enum
import hashlib
import json
import pathlib
import sys
import typing

import jinja2

from .. import exceptions
from .. import helpers
from .. import models_file as models_file_module
from .. import schemas as schemas_module
from .. import types

# TODO(rgreinho)): Fix above nosec it in the future when  # pylint: disable=W0511
# following issue is resolved:
# https://github.com/PyCQA/bandit/issues/211


_DIRECTORY = pathlib.Path(__file__).parent.absolute()
with open(_DIRECTORY / "setup.j2") as in_file:
    _SETUP_TEMPLATE = in_file.read()
with open(_DIRECTORY / "MANIFEST.j2") as in_file:
    _MANIFEST_TEMPLATE = in_file.read()
with open(_DIRECTORY / "init_init_open_alchemy.j2") as in_file:
    _INIT_INIT_OPEN_ALCHEMY_TEMPLATE = in_file.read()
with open(_DIRECTORY / "init.j2") as in_file:
    _INIT_TEMPLATE = in_file.read()


class PackageFormat(enum.Flag):
    """Define the available package formats for the build."""

    NONE = enum.auto()
    SDIST = enum.auto()
    WHEEL = enum.auto()


def validate_dist_format(format_) -> None:
    """
    Ensure a valid package format is provided.

    Raises a ``BuildError` exception if the format is invalid.

    Args:
        format_: the package format to validate
    """
    if format_ is not PackageFormat.NONE and PackageFormat.NONE in format_:
        raise exceptions.BuildError(
            "Combining `PackageFormat.NONE` with another package type "
            "prevents OpenAlchemy to determine if an archive should be "
            f"created or not: '{format_}'. Either specify 'PackageFormat.NONE' "
            "or the format(s) to build. "
        )


def get_schemas(*, spec: typing.Any) -> types.Schemas:
    """
    Get the schemas from the specification.

    Raises MalformedSchemaError if keys to the schemas are missing or the
    schemas are not valid.

    Args:
        spec: The spec to retrieve schemas from.

    Returns:
        The schemas after validation.

    """
    # Check to schemas
    result = schemas_module.validation.spec_validation.check(spec=spec)
    if not result.valid:
        raise exceptions.MalformedSchemaError(result.reason)

    # Check that there is at least 1 model
    assert isinstance(spec, dict)
    components = spec.get("components")
    assert isinstance(components, dict)
    schemas = components.get("schemas")
    assert isinstance(schemas, dict)
    one_model_result = schemas_module.validation.check_one_model(schemas=schemas)
    if not one_model_result.valid:
        raise exceptions.MalformedSchemaError(one_model_result.reason)

    # Check schemas
    schemas_module.process(schemas=schemas)

    return schemas


TVersion = str
TTitle = typing.Optional[str]
TDescription = typing.Optional[str]
TSpecStr = str


def generate_spec_str(
    *,
    schemas: types.Schemas,
    version: TVersion,
    title: TTitle,
    description: TDescription,
) -> TSpecStr:
    """
    Generate a string representation of the spec.

    Args:
        schemas: The schemas of the spec.
        version: The version of the spec.
        title: The title of the spec.
        description: The description of the spec.

    Returns:
        The JSON encoded schemas.

    """
    info = {"version": version}
    if title is not None:
        info["title"] = title
    if description is not None:
        info["description"] = description

    return json.dumps(
        {"info": info, "components": {"schemas": schemas}}, separators=(",", ":")
    )


def calculate_version(*, spec: typing.Any, schemas: types.Schemas) -> TVersion:
    """
    Calculate the version for a spec.

    The algorithm is:
    1. look for the API version in the spec and return or
    2. return the hash of the spec.

    Args:
        spec: The spec as a dictionary.
        spec_str: The spec after it has been converted to string.

    Returns:
        The version of the spec.

    """
    try:
        spec_version = spec["info"]["version"]
        if isinstance(spec_version, TVersion):
            return spec_version
    except (KeyError, TypeError):
        pass

    spec_str = json.dumps({"components": {"schemas": schemas}}, separators=(",", ":"))
    return hashlib.sha1(spec_str.encode()).hexdigest()[:20]


@dataclasses.dataclass
class TSpecInfo:
    """
    Information about the spec.

    Attrs:
        version: Unique identifier for this instance.
        spec_str: The string representation.
        title: A name describing the spec.
        description: More detailed summary of the spec.

    """

    version: TVersion
    spec_str: TSpecStr
    title: TTitle
    description: TDescription


def calculate_spec_info(*, schemas: types.Schemas, spec: typing.Any) -> TSpecInfo:
    """
    Calculate information about the spec.

    Args:
        schemas: The schemas from the spec.
        spec: The spec as a dictionary.

    Returns:
        The spec string to be stored, the version, and the title and description (if
        they are defined).

    """
    version = calculate_version(spec=spec, schemas=schemas)

    title: TTitle = None
    description: TDescription = None
    if "info" in spec:
        title = spec["info"].get("title")
        description = spec["info"].get("description")

    spec_str = generate_spec_str(
        schemas=schemas, version=version, title=title, description=description
    )

    return TSpecInfo(
        version=version, spec_str=spec_str, title=title, description=description
    )


TName = str


def generate_setup(*, name: TName, version: TVersion) -> str:
    """
    Generate the content of the setup.py file.

    Args:
        name: The name of the package.
        version: The version of the package.

    Returns:
        The contents of the setup.py file for the models package.

    """
    template = jinja2.Template(_SETUP_TEMPLATE)

    return template.render(
        name=name,
        version=version,
    )


def generate_manifest(*, name: TName) -> str:
    """
    Generate the content of the MANIFEST.in file.

    Args:
        name: The name of the package.

    Returns:
        The contents of the MANIFEST.in file for the models package.

    """
    template = jinja2.Template(_MANIFEST_TEMPLATE)

    return template.render(
        name=name,
    )


def generate_init_open_alchemy() -> str:
    """
    Generate the OpenAlchemy initialization component of the __init__ file.

    Returns:
        The OpenAlchemy initialization portion of the __init__ file.

    """
    template = jinja2.Template(_INIT_INIT_OPEN_ALCHEMY_TEMPLATE)

    return template.render()


def generate_init_models_file(schemas: types.Schemas) -> str:
    """
    Generate the models file component of init.

    Args:
        schemas: All defined schemas.

    Returns:
        The models file component of the schemas.

    """
    schemas_module.backref.process(schemas=schemas)
    artifacts = schemas_module.artifacts.get_from_schemas(
        schemas=schemas, stay_within_model=False
    )
    return models_file_module.generate(artifacts=artifacts)


def generate_init(open_alchemy: str, models_file: str) -> str:
    """
    Generate the contents for the __init__.py file.

    Args:
        open_alchemy: The OpenAlchemy portion of the __init__ file.
        models_file: The models file portion of the __init__ file.

    Returns:
        The contents of the __init__ file.

    """
    template = jinja2.Template(_INIT_TEMPLATE)

    return template.render(
        open_alchemy=open_alchemy,
        models_file=models_file,
    )


TPath = str


def dump(
    *,
    path: TPath,
    name: TName,
    setup: str,
    manifest: str,
    spec_str: TSpecStr,
    init: str,
) -> None:
    """
    Dump the files needed for the package at a path.

    Args:
        path: The path that will be the root of the package.
        name: The name of the package.
        setup: The contents for the setup file.
        manifest: The contents for the manifest file.
        spec: The contents for the spec file.
        init: The contents for the __init__ file.

    """
    try:
        # Write files at the root of the project directory.
        directory = pathlib.Path(path) / name
        directory.mkdir(parents=True, exist_ok=True)
        (directory / "setup.py").write_text(setup)
        (directory / "MANIFEST.in").write_text(manifest)

        # Write files in the package directory.
        package = directory / name
        package.mkdir(parents=True, exist_ok=True)
        (package / "spec.json").write_text(spec_str)
        (package / "__init__.py").write_text(init)
    except OSError as exc:
        raise exceptions.BuildError(str(exc)) from exc


def build_dist(name: TName, path: TPath, format_: PackageFormat) -> None:
    """
    Build a distribution package.

    The formats can be combined with the bitwise operator "or" (``|``), for
    instance, building both sdist and wheel packages can be specified like this:

    .. code-block:: python

        format_=PackageFormat.SDIST|PackageFormat.WHEEL

    Args:
        path: The package directory.
        format: The package format to build.
    """
    if PackageFormat.SDIST in format_:
        build_sdist(name, path)
    if PackageFormat.WHEEL in format_:
        build_wheel(name, path)


def build_sdist(name: TName, path: TPath) -> None:
    """
    Build a .tar.gz source distribution and place it in a "dist" folder.

    The "dist" folder is located at the root of the project.

    Args:
        path: The package directory.
    """
    pkg_dir = pathlib.Path(path) / name
    helpers.command.run([sys.executable, "setup.py", "sdist"], str(pkg_dir))


def build_wheel(name: TName, path: TPath) -> None:
    """
    Build a .whl package and place it in a "dist" folder.

    The "dist" folder is located at the root of the project.

    Args:
        path: The package directory.
    """
    pkg_dir = pathlib.Path(path) / name
    try:
        helpers.command.run([sys.executable, "setup.py", "bdist_wheel"], str(pkg_dir))
    except exceptions.BuildError as exc:
        raise RuntimeError(
            "Building a wheel package requires the wheel package. "
            "Try `pip install wheel`."
        ) from exc


def execute(
    *,
    spec: typing.Any,
    name: TName,
    path: TPath,
    format_: PackageFormat,
) -> None:
    """
    Execute the build for a spec.

    Args:
        spec: The spec to execute the build on.
        name: The name of the package.
        path: The build output path.
        format_: The format of the distribution package to build.

    """
    validate_dist_format(format_)
    schemas = get_schemas(spec=spec)
    spec_info = calculate_spec_info(schemas=schemas, spec=spec)
    setup = generate_setup(name=name, version=spec_info.version)
    manifest = generate_manifest(name=name)

    init_open_alchemy = generate_init_open_alchemy()
    init_models_file = generate_init_models_file(schemas=schemas)
    init = generate_init(open_alchemy=init_open_alchemy, models_file=init_models_file)

    dump(
        path=path,
        name=name,
        setup=setup,
        manifest=manifest,
        spec_str=spec_info.spec_str,
        init=init,
    )

    # Build a distributable archive if needed.
    build_dist(name, path, format_)



"""Execute external commands."""

import subprocess  # nosec: we are aware of the implications.
import typing

from .. import exceptions


def run(cmd: typing.List[str], cwd: str) -> typing.Tuple[str, str]:
    """
    Run a shell command.

    Args:
        cmd: The command to execute.
        cwd: The path where the command must be executed from.

    Returns:
        A tuple containing (stdout, stderr).

    """
    output = None
    try:
        # "nosec" is used here as we believe we followed the guidelines to use
        # subprocess securely:
        # https://security.openstack.org/guidelines/dg_use-subprocess-securely.html
        output = subprocess.run(  # nosec
            cmd,
            cwd=cwd,
            check=True,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as exc:
        raise exceptions.BuildError(str(exc)) from exc

    return output.stdout.decode("utf-8"), output.stderr.decode("utf-8")



"""Convert from a dictionary to a column value."""

import typing

from ... import exceptions
from ... import helpers
from ... import types as oa_types
from .. import types
from . import array
from . import object_
from . import simple


def convert(*, schema: oa_types.Schema, value: typing.Any) -> types.TAnyCol:
    """
    Convert value for a schema to a dictionary.

    Args:
        value: The value to convert.
        schema: The schema of the value.

    Returns:
        The converted value.

    """
    type_ = helpers.peek.type_(schema=schema, schemas={})
    read_only = helpers.peek.read_only(schema=schema, schemas={})
    if read_only:
        raise exceptions.MalformedModelDictionaryError(
            "readOnly properties cannot be passed to the from_dict constructor."
        )
    json = helpers.peek.json(schema=schema, schemas={})
    if json:
        return value
    if type_ == "object":
        return object_.convert(value, schema=schema)
    if type_ == "array":
        return array.convert(value, schema=schema)
    if type_ in helpers.type_.SIMPLE_TYPES:
        return simple.convert(value, schema=schema)
    raise exceptions.FeatureNotImplementedError(f"Type {type_} is not supported.")



"""All custom exceptions that OpenAlchemy can raise."""


class BaseError(Exception):
    """All exceptions derive at least from this exception."""

    def __init__(self, message, **kwargs):
        """Construct."""
        for key, value in kwargs.items():
            setattr(self, key, value)
        super().__init__(message)

    def __str__(self):
        """Convert exception to string."""
        kwargs_str = ", ".join(f"{key}={value}" for key, value in self.__dict__.items())
        return f"{super().__str__()} {kwargs_str}"


class MissingArgumentError(BaseError, ValueError):
    """Raised when a required argument was not passed."""


class SchemaNotFoundError(BaseError, KeyError):
    """Raised when a schema was not found in the schemas."""


class TypeMissingError(TypeError, BaseError):
    """Raised when a column schema does not have a type."""


class FeatureNotImplementedError(BaseError, NotImplementedError):
    """Raised when a requested feature has not been implemented yet."""


class MalformedSpecificationError(BaseError, ValueError):
    """Raised when an object specification is missing required properties."""


class MalformedSchemaError(BaseError, ValueError):
    """Raised when an object schema is missing required properties."""


class MalformedRelationshipError(BaseError, ValueError):
    """Raised when a relationship was not defined as expected."""


class MalformedExtensionPropertyError(BaseError, ValueError):
    """Raised when an extension property does not have the expected schema."""


class ModelAttributeError(BaseError, AttributeError):
    """Raised when a model lacks a required property."""


class InvalidModelInstanceError(BaseError, ValueError):
    """Raised when model instance object property does not implement to_dict."""


class InvalidInstanceError(BaseError, ValueError):
    """Raised when the an instance of a schema does not conform to the schema."""


class MalformedModelDictionaryError(BaseError, ValueError):
    """Raised when an dictionary of a model does not satisfy the model schema."""


class InheritanceError(BaseError):
    """Raised when an error related to inheritance occurs."""


class BuildError(BaseError):
    """Raised when an error related to build occurs."""


class CLIError(BaseError):
    """Raised when an error occurs when the CLI is used."""



"""Define validation rules for relationship property schema."""

import typing

from ..... import exceptions
from ..... import helpers
from ..... import types as oa_types
from ... import types


def _check_type(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check whether the type of the property is an object or array."""
    # Check type
    try:
        type_ = helpers.peek.type_(schema=schema, schemas=schemas)
        # Check type value
        if type_ not in {"object", "array"}:
            return types.Result(False, "type not an object nor array")
        # Check for JSON
        if (
            helpers.peek.prefer_local(
                get_value=helpers.peek.json, schema=schema, schemas=schemas
            )
            is True
        ):
            return types.Result(False, "property is JSON")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")
    except (exceptions.MalformedSchemaError, exceptions.TypeMissingError) as exc:
        return types.Result(False, f"malformed schema :: {exc}")

    return None


def _check_object_ref(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """
    Check whether a $ref is present and points to a constructable object.

    Assume the type of schema is object.

    """
    # Check for $ref
    ref = helpers.peek.ref(schema=schema, schemas=schemas)
    if ref is None:
        return types.Result(False, "not a reference to another object")

    # Check referenced schema is constructable
    _, ref_schema = helpers.ref.get_ref(ref=ref, schemas=schemas)
    if not helpers.schema.constructable(schema=ref_schema, schemas=schemas):
        return types.Result(False, "referenced schema not constructable")

    return None


def _check_object_backref_uselist(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check backref and uselist for an object."""
    # Check backref
    backref = helpers.peek.prefer_local(
        get_value=helpers.peek.backref, schema=schema, schemas=schemas
    )
    # Check uselist
    uselist = helpers.peek.prefer_local(
        get_value=helpers.peek.uselist, schema=schema, schemas=schemas
    )
    if uselist is False and backref is None:
        return types.Result(
            False, "a one-to-one relationship must define a back reference"
        )

    return None


def _check_kwargs(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check the value of x-kwargs."""
    kwargs = helpers.peek.prefer_local(
        get_value=helpers.peek.kwargs, schema=schema, schemas=schemas
    )
    # Check for unexpected keys
    if kwargs is not None:
        unexpected_keys = {"backref", "secondary"}
        intersection = unexpected_keys.intersection(kwargs.keys())
        if intersection:
            return types.Result(
                False, f"x-kwargs may not contain the {next(iter(intersection))} key"
            )

    return None


def _check_object_values(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check the values of the relationship."""
    # Check nullable
    helpers.peek.prefer_local(
        get_value=helpers.peek.nullable, schema=schema, schemas=schemas
    )
    # Check description
    helpers.peek.prefer_local(
        get_value=helpers.peek.description, schema=schema, schemas=schemas
    )
    # Check writeOnly
    helpers.peek.prefer_local(
        get_value=helpers.peek.write_only, schema=schema, schemas=schemas
    )
    # Check backref and uselist
    backref_uselist_result = _check_object_backref_uselist(
        schema=schema, schemas=schemas
    )
    if backref_uselist_result is not None:
        return backref_uselist_result
    # Check foreign-key-column
    helpers.peek.prefer_local(
        get_value=helpers.peek.foreign_key_column, schema=schema, schemas=schemas
    )
    # Check kwargs
    kwargs_result = _check_kwargs(schema=schema, schemas=schemas)
    if kwargs_result is not None:
        return kwargs_result

    return None


def _check_all_of_duplicates(*, schema: oa_types.Schema) -> types.OptResult:
    """Check for duplicate keys in allOf."""
    # Retrieve allOf
    all_of = schema.get("allOf")
    if all_of is None:
        return None

    # Check for duplicate keys
    seen_keys: typing.Set[str] = set()
    for sub_schema in all_of:
        # Check whether any keys have already been seen
        sub_schema_keys = sub_schema.keys()
        intersection = seen_keys.intersection(sub_schema_keys)
        if intersection:
            return types.Result(
                False, f"multiple {next(iter(intersection))} defined in allOf"
            )

        # Add new keys into seen keys
        seen_keys = seen_keys.union(sub_schema_keys)

    return None


def _check_object(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.Result:
    """Check object property schema."""
    try:
        # Check $ref
        ref_result = _check_object_ref(schema=schema, schemas=schemas)
        if ref_result is not None:
            return ref_result

        # Check nullable
        values_result = _check_object_values(schema=schema, schemas=schemas)
        if values_result is not None:
            return values_result

        # Check for duplicate keys in allOf
        all_of_duplicates_result = _check_all_of_duplicates(schema=schema)
        if all_of_duplicates_result is not None:
            return all_of_duplicates_result

    except exceptions.MalformedSchemaError as exc:
        return types.Result(False, f"malformed schema :: {exc}")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")

    return types.Result(True, None)


def _check_array_root(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check for invalid keys at the array schema root."""
    # Check description
    helpers.peek.description(schema=schema, schemas=schemas)
    # Check writeOnly
    helpers.peek.write_only(schema=schema, schemas=schemas)

    # Check secondary
    if (
        helpers.peek.peek_key(schema=schema, schemas=schemas, key="x-secondary")
        is not None
    ):
        return types.Result(
            False,
            "x-secondary cannot be defined on x-to-many relationship property root",
        )
    # Check backref
    if (
        helpers.peek.peek_key(schema=schema, schemas=schemas, key="x-backref")
        is not None
    ):
        return types.Result(
            False,
            "x-backref cannot be defined on x-to-many relationship property root",
        )
    # Check foreign-key-column
    if (
        helpers.peek.peek_key(
            schema=schema, schemas=schemas, key="x-foreign-key-column"
        )
        is not None
    ):
        return types.Result(
            False,
            "x-foreign-key-column cannot be defined on x-to-many relationship "
            "property root",
        )
    # Check kwargs
    if (
        helpers.peek.peek_key(schema=schema, schemas=schemas, key="x-kwargs")
        is not None
    ):
        return types.Result(
            False,
            "x-kwargs cannot be defined on x-to-many relationship property root",
        )
    # Check uselist
    if (
        helpers.peek.peek_key(schema=schema, schemas=schemas, key="x-uselist")
        is not None
    ):
        return types.Result(
            False,
            "x-uselist cannot be defined on x-to-many relationship property root",
        )

    return None


def _check_array_items_values(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check individual values of array items."""
    # Check items nullable
    items_nullable = helpers.peek.nullable(schema=schema, schemas=schemas)
    if items_nullable is True:
        return types.Result(False, "x-to-many relationships are not nullable")

    # Check items uselist
    items_uselist = helpers.peek.uselist(schema=schema, schemas=schemas)
    if items_uselist is False:
        return types.Result(
            False, "x-to-many relationships do not support x-uselist False"
        )

    return None


def _check_many_to_many(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check many to many schema."""
    # Check items secondary
    secondary = helpers.peek.secondary(schema=schema, schemas=schemas)
    if secondary is None:
        return None

    # Check for foreign key column
    if helpers.peek.foreign_key_column(schema=schema, schemas=schemas) is not None:
        return types.Result(
            False, "many-to-many relationship does not support x-foreign-key-column"
        )

    return None


def _check_array_items(
    *, schema: oa_types.Schema, schemas: oa_types.Schemas
) -> types.OptResult:
    """Check the items schema."""
    # Check items type
    type_result = _check_type(schema=schema, schemas=schemas)
    if type_result is not None:
        return types.Result(
            type_result.valid,
            f"items property :: {type_result.reason}".replace(" nor array", ""),
        )
    type_ = helpers.peek.type_(schema=schema, schemas=schemas)
    if type_ != "object":
        return types.Result(
            False,
            "items property :: type not an object",
        )

    # Check array item values
    _values_result = _check_array_items_values(schema=schema, schemas=schemas)
    if _values_result is not None:
        return _values_result

    # Check items as object
    object_result = _check_object(schema=schema, schemas=schemas)
    if object_result.valid is False:
        return types.Result(
            object_result.valid, f"items property :: {object_result.reason}"
        )

    # Check many to many relationship
    m2m_result = _check_many_to_many(schema=schema, schemas=schemas)
    if m2m_result is not None:
        return m2m_result

    return None


def _check_array(*, schema: oa_types.Schema, schemas: oa_types.Schemas) -> types.Result:
    """Check object property schema."""
    try:
        # Check root schema
        root_result = _check_array_root(schema=schema, schemas=schemas)
        if root_result is not None:
            return root_result

        # Retrieve items schema
        items_schema = helpers.peek.items(schema=schema, schemas=schemas)
        if items_schema is None:
            return types.Result(
                False, "array type properties must define the items schema"
            )

        # Check items
        items_result = _check_array_items(schema=items_schema, schemas=schemas)
        if items_result is not None:
            return items_result

    except exceptions.MalformedSchemaError as exc:
        return types.Result(False, f"malformed schema :: {exc}")
    except exceptions.SchemaNotFoundError as exc:
        return types.Result(False, f"reference :: {exc}")

    return types.Result(True, None)


def check(schemas: oa_types.Schemas, schema: oa_types.Schema) -> types.Result:
    """
    Check whether a property schema is a valid relationship schema.

    At a high level:
    1. the type of the property must be an array or object and not JSON,
    2. the property must reference a constructable schema (at the root for x-to-one and
        at the items level for x-to-many relationships),
    3. any parameters configuring the relationship must have the expected type,
    4. in the allOf list no repeated keys are allowed,
    5. one-to-one relationship must define both x-uselist and x-backref,
    6. x-uselist and nullable is not allowed on x-to-many relationships,
    7. x-foreign-key-column is not allowed on many-to-many relationships,
    8. x-kwargs is not allowed to define backref and secondary and
    9. x-to-many relationships must define property configuration keys at the items
        level.

    Args:
        schemas: All the defined schemas used to resolve any $ref.
        schema: The schema to check.

    Returns:
        Whether the schema is a valid relationship.

    """
    # Check type
    type_result = _check_type(schema=schema, schemas=schemas)
    if type_result is not None:
        return type_result

    # Handle object
    type_ = helpers.peek.type_(schema=schema, schemas=schemas)
    if type_ == "object":
        return _check_object(schema=schema, schemas=schemas)

    # Handle array
    return _check_array(schema=schema, schemas=schemas)



"""Check the schemas are valid."""

import typing

from . import types


def check(*, schemas: typing.Any) -> types.Result:
    """
    Check that the schemas are valid.

    Algorithm:
    1. check that it is a dictionary,
    2. check that every key is a string and
    3. check that every value is a dictionary.

    Args:
        schemas: The schemas to check.

    Returns:
        Whether the schemas are valid and the reason if they are not.

    """
    if not isinstance(schemas, dict):
        return types.Result(False, "schemas must be a dictionary")

    # Check keys are strings
    first_key_not_string = next(
        filter(lambda key: not isinstance(key, str), schemas.keys()),
        None,
    )
    if first_key_not_string is not None:
        return types.Result(
            False, f"schemas keys must be strings, {first_key_not_string} is not"
        )

    # Check values are dictionaries
    first_item_not_dict_value = next(
        filter(lambda args: not isinstance(args[1], dict), schemas.items()), None
    )
    if first_item_not_dict_value is not None:
        key, _ = first_item_not_dict_value
        return types.Result(False, f"the value of {key} must be a dictionary")

    return types.Result(True, None)
